package ac.essex.ooechs.imaging.commons.texture;

import ac.essex.ooechs.imaging.commons.PixelLoader;
import ac.essex.ooechs.imaging.commons.Pixel;
import ac.essex.ooechs.imaging.commons.fast.FastStatistics;

import java.util.ArrayList;

/**
 * Provides statistics about a single-pixel-thickness set of pixels
 * which are a certain radius from the central pixel. The statistics are:
 * mean, standard deviation and edge count. Because this set of pixels
 * is the circular, they will provide the same output even when the image
 * is rotated, which makes it useful for detecting a set of objects which are not
 * all facing the same way.
 * <p/>
 * <p/>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version,
 * provided that any use properly credits the author.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details at http://www.gnu.org
 * </p>
 *
 * // TODO - Also want to calculate edgecount
 *
 * @author Olly Oechsle, University of Essex, Date: 04-Jul-2007
 * @version 1.1
 */
public class Perimeter implements AbstractShape {

    public double[][] mean;
    public double[][] stddev;
    public int[] edgeCount;

    protected ArrayList<Pixel> mask;
    protected int[] maskX, maskY;
    protected int c;

    public Perimeter(int radius) {

/*
        System.out.println("Perimeter: radius=" + radius);

        // calculate masks for each radius
        int circleWidth = radius + radius + 1;
        int perimeterLength = (int) (7 * radius); // AKA radius times 2 pi

        // the points along the perimeter
        maskX = new int[perimeterLength];
        maskY = new int[perimeterLength];
        c = 0;

        for (int y = 0; y < circleWidth; y++) {
            for (int x = 0; x < circleWidth; x++) {

                int dx = x - radius;
                int dy = y - radius;

                double dist = Math.sqrt((dx * dx) + (dy * dy));

                double diff = dist - radius;

                double T = 0.5;

                if (diff < T && diff > -T) {
                        maskX[c] = dx;
                        maskY[c] = dy;
                        c++;
                }

            }
        }
*/

        // Discretised but faster...
        int steps = Math.min(radius, 50);

        maskX = new int[steps];
        maskY = new int[steps];

        c = 0;

        double angleStep = (Math.PI * 2) / steps;
        for (int i = 0; i < steps; i++) {
            double angle = angleStep * i; // two pi radians in a circle
            int dx = (int) (radius * Math.cos(angle));
            int dy = (int) (radius * Math.sin(angle));
            maskX[c] = dx;
            maskY[c] = dy;
            c++;
        }

    }

    public FastStatistics getStatistics(PixelLoader image, int x, int y) {

        FastStatistics statistics = new FastStatistics();

        int width = image.getWidth() - 1;
        int height = image.getHeight() - 1;

        for (int i = 0; i < c; i++) {

            int px = x + maskX[i];

            if (px < 0) continue;
            if (px >= width - 1) continue;

            int py = y + maskY[i];

            if (py < 0) continue;
            if (py >= height - 1) continue;

            statistics.addData(image.getGreyValue(px, py));

        }

        return statistics;

    }

    public float getMean(PixelLoader image, int x, int y) {

        float total = 0;
        int n = 0;

        int width = image.getWidth() - 1;
        int height = image.getHeight() - 1;

        for (int i = 0; i < c; i++) {

            int px = x + maskX[i];

            if (px < 0) continue;
            if (px >= width - 1) continue;

            int py = y + maskY[i];

            if (py < 0) continue;
            if (py >= height - 1) continue;

            total += image.getGreyValue(px, py);
            n++;

        }

        return total / n;

    }


}
