package ac.essex.ooechs.imaging.commons.window.TheMondayProject;

import ac.essex.ooechs.imaging.commons.PixelLoader;
import ac.essex.ooechs.imaging.commons.fast.FastStatistics;
import ac.essex.ooechs.imaging.commons.window.util.WindowFeatures;
import ac.essex.ooechs.imaging.commons.window.TheMondayProject.trends.ChangeTrend;
import ac.essex.ooechs.imaging.commons.window.TheMondayProject.trends.Trend;
import ac.essex.ooechs.imaging.commons.window.TheMondayProject.trends.RangeTrend;
import ac.essex.ooechs.imaging.commons.window.TheMondayProject.trends.StableTrend;

import java.io.File;
import java.util.Vector;
import java.util.Collections;
import java.awt.image.BufferedImage;
import java.awt.*;

/**
 * Monday Project.
 * <p/>
 * A project to identify novelty in a series of images, that is to find
 * whether an image differs signifificantly from the expected value.
 *
 * @author Olly Oechsle, University of Essex, Date: 14-Apr-2008
 * @version 1.0
 */
public class MondayProject {

    public static final int TRUE = 1;
    public static final int FALSE = 2;
    public static final int UNUSUAL = 3;

    Vector<WindowData> data;

    Vector<Trend> trends;

    protected int maxX;

    public static void main(String[] args) throws Exception {

        MondayProject m = new MondayProject();
        m.load("/home/ooechs/Desktop/pipes/Foreground/", TRUE);
        m.load("/home/ooechs/Desktop/pipes/Background/", FALSE);
        m.load("/home/ooechs/Desktop/pipes/Unusual/", UNUSUAL);

        m.print(10, TRUE);
        m.print(10, FALSE);
        m.print(10, UNUSUAL);               

        m.findTrends();
        m.evaluateTrends();
        System.out.println("-------");
        System.out.println("True Data: ");
        m.evaluateData(TRUE);
        System.out.println("-------");
        System.out.println("False Data: ");
        m.evaluateData(FALSE);
        System.out.println("-------");
        System.out.println("Unusual Data: ");
        m.evaluateData(UNUSUAL);

        PixelLoader image = m.createImage(50);
        image.saveAs("/home/ooechs/Desktop/pipes/ranked.png");

    }

    public MondayProject() {
        trends = new Vector<Trend>();
        data = new Vector<WindowData>();
    }

    public Vector<WindowData> getData(int classID) {
        Vector<WindowData> dataWithClassID = new Vector<WindowData>();
        for (int i = 0; i < data.size(); i++) {
            WindowData windowData =  data.elementAt(i);
            if (windowData.classID == classID) dataWithClassID.add(windowData);
        }
        return dataWithClassID;
    }

    public void print(int numToPrint, int classID) {
        Vector<WindowData> dataWithClassID = getData(classID);
        for (int i = 0; i < numToPrint; i++) {
            dataWithClassID.elementAt((int) (dataWithClassID.size() * Math.random())).print();
        }
    }

    public void load(String directoryPath, int classID) {

        File directory = new File(directoryPath);
        if (directory.exists() && directory.isDirectory()) {

            File[] files = directory.listFiles();

            for (int i = 0; i < files.length; i++) {
                File imageFile = files[i];
                try {
                    PixelLoader image = new PixelLoader(imageFile);
                    ac.essex.ooechs.imaging.commons.window.data.Window window = new ac.essex.ooechs.imaging.commons.window.data.Window(image.getWidth(), image.getHeight(), 0, 0, null);
                    double[] values = WindowFeatures.getFeaturesForMonday(image, window);

                    data.add(new WindowData(imageFile.getAbsolutePath(), values, classID));

                    maxX = values.length;

                } catch (Exception err) {
                    // don't worry too much
                }
            }

        }

    }

    /**
     * Searches for trends in the training data.
     */
    public void findTrends() {

        // go through each value of x
        for (int x = 0; x < maxX; x++) {
            // the amount of time for the trend
            for (int S = 0; S < 5; S++) {

                try {

                    // StatisticsObject to find the average value
                    FastStatistics values = new FastStatistics();

                    // Statistics object to find the average change
                    FastStatistics changes = new FastStatistics();

                    // now calculate the average change during that time
                    for (int i = 0; i < data.size(); i++) {

                        WindowData w = data.elementAt(i);

                        if (w.classID != TRUE) continue;

                        double[] features = w.data;

                        for (int s = 0; s < S; s++) {
                            values.addData((float) features[x + s]);
                        }

                        double value1 = features[x];
                        double value2 = features[x + S];

                        double change = (value2 / value1) - 1;

                        changes.addData((float) change);

                    }

                    RangeTrend rangeTrend = new RangeTrend(x, S, values.getMin(), values.getMax());

                    addTrend(rangeTrend);

                    double changesStdDev = changes.getStandardDeviation();
                    double changesMean = changes.getMean();

                    double lowestScore;

                    lowestScore = Double.MAX_VALUE;
                    ChangeTrend bestChangeTrend = null;

                    for (double tolerance = 0; tolerance < changesStdDev * 2; tolerance+= (changesStdDev / 10))  {
                        ChangeTrend changeTrend = new ChangeTrend(x, S, changesMean, tolerance);
                        double score = getScore(changeTrend);
                        if (score < lowestScore) {
                            lowestScore = score;
                            bestChangeTrend = changeTrend;
                        }

                    }

                    if (bestChangeTrend != null && Math.abs(bestChangeTrend.minChange) >= 0.05) {
                        addTrend(bestChangeTrend);
                    }

                    StableTrend bestStableTrend = null;

                    for (double tolerance = 0; tolerance < changesStdDev * 2; tolerance+= (changesStdDev / 10))  {
                        StableTrend stableTrend = new StableTrend(x, S, changesMean, tolerance);
                        double score = getScore(stableTrend);
                        if (score < lowestScore) {
                            lowestScore = score;
                            bestStableTrend = stableTrend;
                        }

                    }

                    if (bestStableTrend != null) {
                        addTrend(bestStableTrend);
                    }



                } catch (ArrayIndexOutOfBoundsException e) {


                }

            }
        }
    }

    public void addTrend(Trend t) {
        double trueScore = getMatchProbability(t, TRUE);
        double falseScore = getMatchProbability(t, FALSE);
        if (trueScore > 0.8 && falseScore < 0.2) trends.add(t);
    }

    public double getScore(Trend t) {
        double trueScore = getMatchProbability(t, TRUE);
        double falseScore = getMatchProbability(t, FALSE);
        return (1 - trueScore) + falseScore;
    }

    public void evaluateTrends() {
        for (int i = 0; i < trends.size(); i++) {
            Trend trend = trends.elementAt(i);
            double trueMatchProbability = getMatchProbability(trend, TRUE);
            double falseMathProbability = getMatchProbability(trend, FALSE);
            trend.trueMatchProbability = trueMatchProbability;
            System.out.println(i + ": " + trueMatchProbability + ", " + falseMathProbability);
        }
    }

    public void evaluateData(int classID) {
        FastStatistics stats = new FastStatistics();
        for (int i = 0; i < data.size(); i++) {

            WindowData w = data.elementAt(i);

            if (w.classID != classID) continue;

            double[] features = w.data;

            double matchCount = 0;
            double failureCount = 0;

            // hist how many of the trends match this particular value
            for (int j = 0; j < trends.size(); j++) {
                Trend trend = trends.elementAt(j);
                if (trend.matches(features)) {
                    matchCount++;
                } else {
                    failureCount += trend.trueMatchProbability;
                }
            }

            stats.addData((float) matchCount);

            w.rank = matchCount;

        }

        System.out.println("Total trends: " + trends.size());
        System.out.println("Mean matches: " + stats.getMean());
        System.out.println("Std Deviation: " + stats.getStandardDeviation());
        System.out.println("Min: " + stats.getMin());
        System.out.println("Max: " + stats.getMax());

    }

    public PixelLoader createImage(int width) throws Exception {
        int totalHeight = 5 * data.size();
        System.out.println("Creating Image: " + width + "x" + totalHeight);
        BufferedImage img = new BufferedImage(width, totalHeight, BufferedImage.TYPE_INT_ARGB);
        Graphics g = img.getGraphics();
        int y = 0;
        Collections.sort(data);
        for (int i = 0; i < data.size(); i++) {
            WindowData windowData =  data.elementAt(i);
            PixelLoader image = new PixelLoader(windowData.file);
            g.drawImage(image.getBufferedImage(), 0, y, null);
            y+= image.getHeight();
        }
        return new PixelLoader(img);
    }

    public double getMatchProbability(Trend trend, int classID) {

        int matches = 0;
        int total = 0;

        for (int i = 0; i < data.size(); i++) {
            WindowData w = data.elementAt(i);

            if (w.classID != classID) continue;

            double[] features = w.data;

            if (trend.matches(features)) {
                matches++;
            }

            total++;

        }

        return matches / (double) total;

    }


}
